/*
 * kgraft_patch_bsc1096723
 *
 * Fix for CVE-2018-1000026, bsc#1096723
 *
 *  Upstream commits:
 *  2b16f048729b ("net: create skb_gso_validate_mac_len()")
 *  8914a595110a ("bnx2x: disable GSO where gso_size is too big for hardware")
 *
 *  SLE12(-SP1) commits:
 *  none yet
 *
 *  SLE12-SP2 commits:
 *  14f582fc9ad9687bd87dc8bae0363b6b579f0d6c
 *  f6d5e3598463dc3f2e6ba19558f0ecbc4d7cf9ef
 *
 *  SLE12-SP3 commits:
 *  5eb4578abf5a90771c23ab2893c9b1857871f7d3
 *  7bd42f8c29fa0a9f713e1062ebf9bcb20a825f0a
 *
 *  SLE15 commits:
 *  f2e13b50306dd366583f44c03273b0f714f184d5
 *  d647fd13fdfd2f43651342d5e95522691eddca31
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_vlan.h>
#include <net/vxlan.h>
#include "kgr_patch_bsc1096723.h"

#if !IS_MODULE(CONFIG_BNX2X)
#error "Live patch supports only CONFIG_BNX2X=m."
#endif


/* new */
static inline unsigned int kgr_skb_gso_mac_seglen(const struct sk_buff *skb)
{
	unsigned int hdr_len = skb_transport_header(skb) - skb_mac_header(skb);
	return hdr_len + skb_gso_transport_seglen(skb);
}

/* new */
static bool kgr_skb_gso_validate_mac_len(const struct sk_buff *skb, unsigned int len)
{
	return kgr_skb_gso_mac_seglen(skb) <= len;
}

/* patched */
netdev_features_t kgr_bnx2x_features_check(struct sk_buff *skb,
					   struct net_device *dev,
					   netdev_features_t features)
{
	/*
	 * Fix CVE-2018-1000026
	 *  +18 lines
	 */
	/*
	 * A skb with gso_size + header length > 9700 will cause a
	 * firmware panic. Drop GSO support.
	 *
	 * Eventually the upper layer should not pass these packets down.
	 *
	 * For speed, if the gso_size is <= 9000, assume there will
	 * not be 700 bytes of headers and pass it through. Only do a
	 * full (slow) validation if the gso_size is > 9000.
	 *
	 * (Due to the way SKB_BY_FRAGS works this will also do a full
	 * validation in that case.)
	 */
	if (unlikely(skb_is_gso(skb) &&
		     (skb_shinfo(skb)->gso_size > 9000) &&
		     !kgr_skb_gso_validate_mac_len(skb, 9700)))
		features &= ~NETIF_F_GSO_MASK;

	features = vlan_features_check(skb, features);
	return vxlan_features_check(skb, features);
}
